(in-package "NEWCMP")

;;;;
;;;; Register Assignment
;;;;

(defun find-register (v regs)
  (let ((r (cdr (assoc v regs))))
    (unless r (error "can't find register for ~s" v))
    r))

(defstruct (intgr-node (:constructor new-intgr-node (values register)))
  values register index)

(defstruct intgr
  (valmap (make-hash-table)) (edges (make-hash-table)) moves)

(defvar *intgr-node-index*)

(defun make-intgr-node (vals graph &optional register)
  (let ((n (new-intgr-node vals register))
        (valmap (intgr-valmap graph)))
    (setf (intgr-node-index n) *intgr-node-index*)
    (incf *intgr-node-index*)
    (setf (gethash n (intgr-edges graph)) nil)
    (dolist (v vals) (setf (gethash v valmap) n))
    n))

(defun find-intgr-node (v graph)
  (gethash v (intgr-valmap graph)))

(defun add-intgr-edge (i j graph)
  (pushnew j (gethash i (intgr-edges graph)))
  (pushnew i (gethash j (intgr-edges graph))))

(defun intgr-has-edge (i j graph)
  (member j (gethash i (intgr-edges graph))))

(defun intgr-siblings (i gr)
  (gethash i (intgr-edges gr)))

(defun intgr-delete-node (i gr)
  (let* ((edges (intgr-edges gr))
         (sibs (gethash i edges))
         (valmap (intgr-valmap gr)))
    (remhash i edges)
    (dolist (v (intgr-node-values i)) (remhash v valmap))
    (dolist (s sibs)
      (setf (gethash s edges) (delete i (gethash s edges) :count 1)))))

(defun intgr-nodes (gr)
  (let ((nodes nil))
    (maphash #'(lambda (x y) (declare (ignore y)) (push x nodes))
             (intgr-edges gr))
    (stable-sort nodes #'(lambda (x y)
                           (> (intgr-node-index x) (intgr-node-index y))))))

(defun show-intgr (gr)
  (flet ((show-node (n)
           (mapcar #'(lambda (v) (if (var-p v) v 'primop))
                   (intgr-node-values n))))
    (mapcar #'(lambda (n)
                (list (show-node n)
                      (mapcar #'show-node (intgr-siblings n gr))))
            (intgr-nodes gr))))

(defun fill-intgr-cexpr (e env graph)
  (if (letrec-p e)
      (fill-intgr-letrec e env graph)
    (let* ((conts (cexpr-conts e))
           (vars (cexpr-vars e))
           (args (cexpr-args e))
           (nodes (mapcar #'(lambda (v) (make-intgr-node (list v) graph)) vars))
           (cfree (let ((free nil))
                    (dolist (c conts free)
                      (let ((cf (fill-intgr-cexpr c env graph)))
                        (setf free (union cf free))))))
           (outfree (set-difference cfree vars))
           (fargs (let ((fun (cexpr-fun e)))
                    (if (and (var-p fun) (null (assoc fun env)))
                        (list fun)
                      nil)))
           (infree (union (union fargs args) outfree)))
      (do ((nodes nodes (rest nodes)))
          ((null nodes))
          (let ((n (first nodes)))
            (dolist (i (rest nodes))
              (add-intgr-edge n i graph))))
      (dolist (n nodes)
        (dolist (f outfree)
          (let ((fn (find-intgr-node f graph)))
            (when fn (add-intgr-edge n fn graph)))))
      (cond
        ((and (null vars) (primop-p e))
         (let ((s (make-intgr-node (list e) graph)))
           (dolist (f outfree)
             (let ((fn (find-intgr-node f graph)))
               (when fn (add-intgr-edge s fn graph))))))
        ((local-tail-call-p e env)
         (do ((src (call-args e) (rest src))
              (tgt (func-vars (cdr (assoc (call-fun e) env))) (rest tgt)))
             ((or (null src) (null tgt)))
             (push (list (first src) (first tgt) 0) (intgr-moves graph))))
        ((escaping-function-p e)
         (let ((r 1))
           (dolist (n nodes)
             (setf (intgr-node-register n) r)
             (incf r)))))
      infree)))

(defun find-primop-value-register (e regs)
  (let ((vars (primop-vars e)))
    (find-register (if vars (first vars) e) regs)))

(defun local-tail-call-p (e env)
  (and (call-p e) (null (call-form e)) (assoc (call-fun e) env)))

(defun escaping-function-p (e)
  (and (func-p e) (member (func-type e) '(lambda local-lambda))))
(defun fill-intgr-letrec (e env graph)
  (let* ((args (letrec-args e))
         (vars (letrec-vars e))
         (nenv (append (mapcar #'cons vars args) env)))
    (dolist (a args) (fill-intgr-cexpr a nenv graph))
    (fill-intgr-cexpr (letrec-body e) nenv graph)))

(defun intgr-coalesce-nodes (i j gr)
  (when (or (eq i j) (intgr-has-edge i j gr))
        (break)
        (error "can't coalesce nodes that interfere"))
  (let ((i-sibs (intgr-siblings i gr))
        (j-sibs (intgr-siblings j gr))
        (ri (intgr-node-register i))
        (rj (intgr-node-register j))
        (vals (append (intgr-node-values i) (intgr-node-values j))))
    (when (and ri rj (/= ri rj))
          (error "conflicting register assignments"))
    (intgr-delete-node i gr)
    (intgr-delete-node j gr)
    (let ((n (make-intgr-node vals gr (if ri ri rj))))
      (dolist (is i-sibs) (add-intgr-edge n is gr))
      (dolist (js j-sibs) (add-intgr-edge n js gr))
      n)))

(defun coalesce-preassigned-nodes (gr)
  (let ((alist nil))
    (dolist (n (intgr-nodes gr))
      (let ((r (intgr-node-register n)))
        (when r
              (let ((entry (assoc r alist)))
                (if entry
                    (push n (rest entry))
                  (push (list r n) alist))))))
    (dolist (nodes (mapcar #'rest alist))
      (let ((i (first nodes)))
        (dolist (j (rest nodes))
          (setf i (intgr-coalesce-nodes i j gr)))))))

(defun coalesce-intgr-moves (gr)
  (let ((moves (stable-sort (intgr-moves gr)
                            #'(lambda (x y) (> (third x) (third y))))))
    (dolist (m moves)
      (let ((i (find-intgr-node (first m) gr))
            (j (find-intgr-node (second m) gr)))
        (unless (or (eq i j) (intgr-has-edge i j gr))
                (intgr-coalesce-nodes i j gr))))))

(defun intgr-color (gr)
  (dolist (n (intgr-nodes gr))
    (unless (intgr-node-register n)
            (let* ((used (let ((regs nil))
                           (dolist (s (intgr-siblings n gr) regs)
                             (let ((r (intgr-node-register s)))
                               (when r (push r regs))))))
                   (r (do ((i 1 (+ i 1)))
                          ((not (member i used)) i))))
              (setf (intgr-node-register n) r))))
    (let ((regs nil))
      (dolist (n (intgr-nodes gr) regs)
        (let ((r (intgr-node-register n)))
          (dolist (v (intgr-node-values n))
            (push (cons v r) regs))))))

(defun regassign (e)
  (let ((gr (make-intgr))
        (*intgr-node-index* 0))
    (fill-intgr-cexpr e nil gr)
    (coalesce-preassigned-nodes gr)
    (coalesce-intgr-moves gr)
    (intgr-color gr)))


;;;;
;;;; Computing Frame Sizes
;;;;

(defun find-regmax (e rmax regs)
  (if (letrec-p e)
      (find-regmax (letrec-body e) rmax regs)
    (let* ((vregs (mapcar #'(lambda (x) (find-register x regs))
                          (cexpr-vars e)))
           (nrmax (reduce #'max vregs :initial-value rmax)))
      (if (and (primop-p e) (null (cexpr-vars e)))
          (let* ((sreg (find-register e regs))
                 (nnrmax (if sreg (max sreg nrmax) nrmax)))
            (find-regmax-conts (cexpr-conts e) nnrmax regs))
        (find-regmax-conts (cexpr-conts e) nrmax regs)))))

(defun find-regmax-conts (conts rmax regs)
  (if (null conts)
      rmax
    (let ((nrmax (find-regmax (first conts) rmax regs)))
      (find-regmax-conts (rest conts) nrmax regs))))

(defun compute-frame-sizes (e regs)
  (let* ((gr (compute-callgr e))
         (funs (callgr-functions gr))
         (env (mapcar #'(lambda (x) (cons x (find-regmax x 0 regs))) funs)))
    (loop
     (let ((changed nil))
       (dolist (f funs)
         (let ((entry (assoc f env)))
           (dolist (c (callgr-callees f gr))
             (when (eq (func-type c) 'tail-lambda)
                   (let ((cmax (cdr (assoc c env))))
                     (when (< (cdr entry) cmax)
                           (setf (cdr entry) cmax)
                           (setf changed t)))))))
       (unless changed (return))))
    env))

(defun find-frame-size (e env)
  (cdr (assoc e env)))


;;;;
;;;; Code Representation and Emission
;;;;

(defun push-instruction (inst cl)
  (push inst (first cl)))

(defun push-return-instruction (cl)
  (let ((prev (first (first cl))))
    (case (if (consp prev) (first prev) nil)
          (xlscmp::%set-one-value
           (setf (first (first cl))
                 `(xlscmp::%set-one-value-return 0 ,(second prev))))
          (xlscmp::%set-values
           (setf (first (first cl))
                 `(xlscmp::%set-values-return 0 ,@(rest prev))))
          (xlscmp::%set-values-list
           (setf (first (first cl))
                 `(xlscmp::%set-values-list-return 0 ,(second prev))))
          (t (push-instruction `(xlscmp::%return 0) cl)))))

(defun push-label-instruction (label cl)
  (let ((prev (first (first cl))))
    (cond
     ((and (consp prev)
           (eq (first prev) 'xlscmp::%goto)
           (eq (second prev) label))
      (pop (first cl))
      (push-instruction label cl))
     (t (push-instruction label cl)))))

(defun find-literal (lit cl)
  (let ((n (length (second cl)))
        (i (position lit (second cl))))
    (if i
        (- n 1 i)
      (progn (push lit (second cl))
             n))))

(defun make-label-var (sym)
  (make-var sym t nil))


;;;;
;;;; Main Entry Point
;;;;

(defun gencode (e)
  (let* ((mv (make-label-var 'main))
         (cl (list (list mv) nil (list mv)))
         (regs (regassign e))
         (env (compute-frame-sizes e regs)))
    (gencode-cexpr e regs env cl)
    (values (nreverse (first cl))
            (nreverse (second cl))
            (nreverse (third cl)))))

(defun gencode-cexpr (e regs env cl)
  (cond
   ((func-p e)
    (case (func-type e)
          (lambda (gencode-lambda e regs env cl))
          (local-lambda (gencode-local-lambda e regs env cl))
          (tail-lambda (gencode-tail-lambda e regs env cl))))
   ((primop-p e) (gencode-primop e regs env cl))
   ((testop-p e) (gencode-testop e regs env cl))
   ((letrec-p e) (gencode-letrec e regs env cl))
   ((call-p e) (gencode-call e regs env cl))
   (t (error "don't know how to generate code for ~s" e))))


;;;;
;;;; Function Code Generators
;;;;

(defun gencode-lambda (e regs env cl)
  (let* ((nv (length (func-vars e)))
         (nr (find-frame-size e env))
         (ntmp (- nr nv))
         (nreq (func-nreq e))
         (nopt (func-nopt e))
         (have-rest (func-has-rest e)))
    (cond
     (have-rest
      (let* ((nreq+1 (+ nreq 1))
             (odefs (make-array nopt :initial-element 'xlscmp::%not-supplied))
             (ol (find-literal odefs cl)))
        (if (= nopt 0)
            (push-instruction `(xlscmp::%initialize 2 ,nreq+1 0 0 ,ntmp) cl)
          (push-instruction `(xlscmp::%initialize 2 ,nreq+1 ,nopt ,ol 0 ,ntmp)
                            cl))))
     ((< 0 nopt)
      (let* ((nreq+1 (+ nreq 1))
             (odefs (make-array nopt :initial-element 'xlscmp::%not-supplied))
             (ol (find-literal odefs cl)))
        (push-instruction `(xlscmp::%initialize 1 ,nreq+1 ,nopt ,ol 0 ,ntmp)
                          cl)))
     (t (push-instruction `(xlscmp::%initialize-0 ,(+ nv 1) 0 ,ntmp) cl)))
    (gencode-cexpr (func-body e) regs env cl)))

(defun gencode-local-lambda (e regs env cl)
  (gencode-lambda e regs env cl))

(defun gencode-tail-lambda (e regs env cl)
  (gencode-cexpr (func-body e) regs env cl))


;;;;
;;;; Primop Code Generators
;;;;

(let ((table (make-hash-table)))
  (defun get-primop-generator (op) (gethash op table))
  (defun set-primop-generator (op g) (setf (gethash op table) g)))

(defun gencode-primop (e regs env cl)
  (let ((g (get-primop-generator (primop-fun e))))
    (unless g (error "can't generate code for primop ~s" (primop-fun e)))
    (funcall g e regs env cl)))

(defmacro define-primop-generator (which args &rest body)
  `(progn
     (set-primop-generator ',which #'(lambda ,args ,@body))
     ',which))

(defun gencode-primop-cont-1 (e regs env cl)
  (let ((form (primop-form e)))
    (if form
        (gencode-cexpr form regs env cl)
      (let ((vr (find-primop-value-register e regs)))
        (push-instruction `(xlscmp::%set-one-value ,vr) cl)
        (push-return-instruction cl)))))

(defmacro define-simple-primop-generator (which (aregs vreg) form)
  `(define-primop-generator ,which (e regs env cl)
     (let ((,aregs (mapcar #'(lambda (a) (find-register a regs))
                           (primop-args e)))
           (,vreg (find-primop-value-register e regs)))
       (push-instruction ,form cl))
     (gencode-primop-cont-1 e regs env cl)))

(defmacro define-standard-primop-generator (name &optional (symbol name))
  `(define-simple-primop-generator ,name (a v) `(,',symbol ,@a ,v)))

(define-primop-generator load-constant (e regs env cl)
  (let ((ci (find-literal (primop-data e) cl))
        (v (find-primop-value-register e regs)))
    (push-instruction `(xlscmp::%ldconst ,ci ,v) cl)
    (gencode-primop-cont-1 e regs env cl)))

(defmacro define-arith1-generator (sym char)
  `(define-simple-primop-generator ,sym (a v)
     `(xlscmp::%arith1 ,(char-int ,char) ,@a ,v)))

(defmacro define-arith2-generator (sym char)
  `(define-simple-primop-generator ,sym (a v)
     `(xlscmp::%arith2 ,(char-int ,char) ,@a ,v)))

(defmacro define-arith-pred2-generator (sym char)
  `(define-simple-primop-generator ,sym (a v)
     `(xlscmp::%arith-pred2 ,(char-int ,char) ,@a ,v)))

(define-arith2-generator + #\+)
(define-arith2-generator - #\-)
(define-arith2-generator * #\*)
(define-arith2-generator / #\/)
(define-arith2-generator min #\m)
(define-arith2-generator max #\M)

(define-arith1-generator uminus #\-)
(define-arith1-generator udiv #\/)
(define-arith1-generator 1+ #\p)
(define-arith1-generator 1- #\m)
(define-arith-pred2-generator <  #\<)
(define-arith-pred2-generator <= #\L)
(define-arith-pred2-generator =  #\=)
(define-arith-pred2-generator /= #\#)
(define-arith-pred2-generator >= #\G)
(define-arith-pred2-generator >  #\>)

(define-primop-generator values (e regs env cl)
  (let ((aregs (mapcar #'(lambda (a) (find-register a regs)) (primop-args e)))
        (form (primop-form e)))
    (push-instruction (if (= (length aregs) 1)
                          `(xlscmp::%set-one-value ,(first aregs))
                        `(xlscmp::%set-values ,(length aregs) ,@aregs))
                      cl)
    (if form
        (gencode-cexpr form regs env cl)
      (push-return-instruction cl))))

(define-primop-generator values-list (e regs env cl)
  (let ((areg (find-register (first (primop-args e)) regs))
        (form (primop-form e)))
    (push-instruction `(xlscmp::%set-values-list ,areg) cl)
    (if form
        (gencode-cexpr form regs env cl)
      (push-return-instruction cl))))

(define-standard-primop-generator nth-value xlscmp::%nth-value)

(define-standard-primop-generator get-one-value xlscmp::%get-one-value)

(define-primop-generator load-mvargs (e regs env cl)
  (let* ((data (primop-data e))
         (nreq (first data))
         (nopt (second data))
         (has-rest (if (third data) 1 0))
         (ni (find-literal (fourth data) cl))
         (v (mapcar #'(lambda (x) (find-register x regs)) (primop-vars e))))
    (push-instruction `(xlscmp::%ldmvargs ,nreq ,nopt ,has-rest ,ni ,@v) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator save-values (e regs env cl)
  (let ((nr (find-register (first (primop-vars e)) regs)))
    (push-instruction `(xlscmp::%push-values ,nr) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator restore-values (e regs env cl)
  (let ((nr (find-register (first (primop-args e)) regs))
        (form (primop-form e)))
    (push-instruction `(xlscmp::%pop-values ,nr) cl)
    (if form
        (gencode-cexpr (primop-form e) regs env cl)
      (push-return-instruction cl))))

(define-primop-generator block (e regs env cl)
  (let ((label (make-label-var 'end))
        (bi (find-register (first (primop-args e)) regs)))
    (push-instruction `(xlscmp::%new-block ,bi ,label) cl)
    (gencode-catch-cont e label regs env cl)))

(defun gencode-catch-cont (e label regs env cl)
  (let ((form (primop-form e))
        (var (first (primop-vars e))))
    (push-label-instruction label cl)
    (cond
     (form
      (when var
            (let ((vi (find-register var regs)))
              (push-instruction `(xlscmp::%get-one-value ,vi) cl)))
      (gencode-cexpr form regs env cl))
     (t (push-return-instruction cl)))))

(define-primop-generator return-from (e regs env cl)
  (declare (ignore env))
  (let ((ti (find-register (first (primop-args e)) regs)))
    (push-instruction `(xlscmp::%throw-return-from ,ti) cl)))

(define-primop-generator tagbody (e regs env cl)
  (let ((label (make-label-var 'end))
        (bi (find-register (first (primop-args e)) regs)))
    (push-instruction `(xlscmp::%new-tagbody ,bi ,label) cl)
    (gencode-catch-cont e label regs env cl)))

(define-primop-generator go (e regs env cl)
  (declare (ignore env))
  (let* ((args (primop-args e))
         (ti (find-register (first args) regs))
         (label (find-register (second args) regs)))
    (push-instruction `(xlscmp::%new-go ,ti ,label) cl)))

(define-primop-generator catch (e regs env cl)
  (let ((label (make-label-var 'end))
        (ti (find-register (first (primop-args e)) regs))
        (ci (find-register (second (primop-args e)) regs)))
    (push-instruction `(xlscmp::%new-catch ,ti ,ci ,label) cl)
    (gencode-catch-cont e label regs env cl)))

(define-primop-generator throw (e regs env cl)
  (declare (ignore env))
  (let ((ti (find-register (first (primop-args e)) regs)))
    (push-instruction `(xlscmp::%throw ,ti) cl)))

(define-primop-generator errset (e regs env cl)
  (let ((label (make-label-var 'end))
        (bi (find-register (first (primop-args e)) regs))
        (vi (find-register (second (primop-args e)) regs)))
    (push-instruction `(xlscmp::%new-errset ,bi ,vi ,label) cl)
    (gencode-catch-cont e label regs env cl)))

(define-primop-generator unwind-protect (e regs env cl)
  (let ((label (make-label-var 'end))
        (bi (find-register (first (primop-args e)) regs))
        (vi (find-register (second (primop-args e)) regs)))
    (push-instruction `(xlscmp::%new-unwind-protect ,bi ,vi ,label) cl)
    (gencode-catch-cont e label regs env cl)))

(define-primop-generator make-closures (e regs env cl)
  (let* ((vr (mapcar #'(lambda (x) (find-register x regs)) (primop-vars e)))
         (na (primop-data e))
         (nv (length vr)))
    (push-instruction `(xlscmp::%make-closures ,na ,nv ,@vr ) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator set-closure-data (e regs env cl)
  (let* ((args (primop-args e))
         (ci (find-register (first args) regs))
         (ar (mapcar #'(lambda (x) (find-register x regs)) (rest args)))
         (na (length ar)))
    (push-instruction `(xlscmp::%set-closure-data ,ci ,na ,@ar ) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator set-closure-code (e regs env cl)
  (let* ((args (primop-args e))
         (cr (find-register (first args) regs))
         (label (second args))
         (li (find-literal (primop-data e) cl)))
    (push-instruction `(xlscmp::%set-closure-code ,cr ,label ,li) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator dynamic-bind (e regs env cl)
  (let* ((args (primop-args e))
         (sr (find-register (first args) regs))
         (vr (find-register (second args) regs))
         (er (find-register (first (primop-vars e)) regs)))
    (push-instruction `(xlscmp::%new-dynamic-bind ,sr ,vr ,er) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator dynamic-unbind (e regs env cl)
  (let ((form (primop-form e))
        (er (find-register (first (primop-args e)) regs)))
    (push-instruction `(xlscmp::%new-dynamic-unbind ,er) cl)
    (if form
        (gencode-cexpr form regs env cl)
      (push-return-instruction cl))))

(define-primop-generator struct-ref (e regs env cl)
  (let ((si (find-register (first (primop-args e)) regs))
        (i (primop-data e))
        (v (find-primop-value-register e regs)))
    (push-instruction `(xlscmp::%struct-op ,si 0 ,i ,v) cl)
    (gencode-primop-cont-1 e regs env cl)))

(define-primop-generator set-struct-ref (e regs env cl)
  (let ((si (find-register (first (primop-args e)) regs))
        (i (primop-data e))
        (ni (find-register (second (primop-args e)) regs))
        (v (find-primop-value-register e regs)))
    (push-instruction `(xlscmp::%struct-op ,si 1 ,i ,ni ,v) cl)
    (gencode-primop-cont-1 e regs env cl)))

(define-primop-generator struct-type-p (e regs env cl)
  (let ((si (find-register (first (primop-args e)) regs))
        (ti (find-literal (primop-data e) cl))
        (v (find-primop-value-register e regs)))
    (push-instruction `(xlscmp::%struct-op ,si 2 ,ti ,v) cl)
    (gencode-primop-cont-1 e regs env cl)))

(define-primop-generator symbol-value (e regs env cl)
  (let ((si (find-literal (primop-data e) cl))
        (r (find-primop-value-register e regs)))
    (push-instruction `(xlscmp::%symval ,si ,r) cl))
    (gencode-primop-cont-1 e regs env cl))

(define-primop-generator symbol-function (e regs env cl)
  (let ((si (find-literal (primop-data e) cl))
        (r (find-primop-value-register e regs)))
    (push-instruction `(xlscmp::%symfun ,si ,r) cl))
    (gencode-primop-cont-1 e regs env cl))

(define-primop-generator set-symbol-value (e regs env cl)
  (let ((ci (find-literal (primop-data e) cl))
        (v (find-register (first (primop-args e)) regs))
        (r (find-primop-value-register e regs)))
    (push-instruction `(xlscmp::%set-symval ,ci ,v ,r) cl)
    (gencode-primop-cont-1 e regs env cl)))

(define-standard-primop-generator get xlscmp::%get)
(define-standard-primop-generator xlscmp::%set-get)

(define-standard-primop-generator xlscmp::%supplied-p)
(define-standard-primop-generator load-missing xlscmp::%ldnotsupp)

(define-primop-generator get-optarg (e regs env cl)
  (let ((a (find-register (first (primop-args e)) regs))
        (d (find-literal (primop-data e) cl))
        (v (find-register (first (primop-vars e)) regs))
        (s (find-register (second (primop-vars e)) regs)))
    (push-instruction `(xlscmp::%get-optarg ,a ,d ,v, s) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator make-keyargs (e regs env cl)
  (let ((a (find-register (first (primop-args e)) regs))
        (v (find-register (first (primop-vars e)) regs)))
    (push-instruction `(xlscmp::%make-keyargs ,a ,v) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator get-keyarg (e regs env cl)
  (let ((a (find-register (first (primop-args e)) regs))
        (k (find-literal (first (primop-data e)) cl))
        (d (find-literal (second (primop-data e)) cl))
        (v (find-register (first (primop-vars e)) regs))
        (s (find-register (second (primop-vars e)) regs)))
    (push-instruction `(xlscmp::%get-keyarg ,a ,k ,d ,v ,s) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-primop-generator check-last-keyarg (e regs env cl)
  (let ((a (find-register (first (primop-args e)) regs))
        (d (if (primop-data e) 1 0)))    
    (push-instruction `(xlscmp::%check-last-keyarg ,a ,d) cl)
    (gencode-cexpr (primop-form e) regs env cl)))

(define-standard-primop-generator cons xlscmp::%cons)
(define-standard-primop-generator car xlscmp::%car)
(define-standard-primop-generator cdr xlscmp::%cdr)
(define-standard-primop-generator xlscmp::%set-car)
(define-standard-primop-generator xlscmp::%set-cdr)
(define-standard-primop-generator rplaca xlscmp::%rplaca)
(define-standard-primop-generator rplacd xlscmp::%rplacd)

(defmacro define-cxr-generator (name n x)
  `(define-simple-primop-generator ,name (a v)
     (list 'xlscmp::%cxr ,n ,x (first a) v)))

(define-cxr-generator caar 2 #b11)
(define-cxr-generator cadr 2 #b10)
(define-cxr-generator cdar 2 #b01)
(define-cxr-generator cddr 2 #b00)

(define-cxr-generator caaar 3 #b111)
(define-cxr-generator caadr 3 #b110)
(define-cxr-generator cadar 3 #b101)
(define-cxr-generator caddr 3 #b100)
(define-cxr-generator cdaar 3 #b011)
(define-cxr-generator cdadr 3 #b010)
(define-cxr-generator cddar 3 #b001)
(define-cxr-generator cdddr 3 #b000)

(define-standard-primop-generator nth xlscmp::%nth)
(define-standard-primop-generator xlscmp::%set-nth)

(define-simple-primop-generator aref (a v)
  (case (length a)
    (2 `(xlscmp::%aref1 ,@a ,v))
    (3 `(xlscmp::%aref2 ,@a ,v))
    (t (error "wrong argument count for ~s" 'aref))))
  
(define-simple-primop-generator xlscmp::%set-aref (a v)
  (case (length a)
    (3 `(xlscmp::%set-aref1 ,@a ,v))
    (4 `(xlscmp::%set-aref2 ,@a ,v))
    (t (error "wrong argument count for ~s" 'xlscmp::%set-aref))))

(define-standard-primop-generator row-major-aref xlscmp::%aref1)
(define-standard-primop-generator xlisp::%set-rm-aref xlscmp::%set-aref1)
(define-standard-primop-generator svref xlscmp::%svref)
(define-standard-primop-generator xlscmp::%set-svref)

(define-standard-primop-generator make-cell xlscmp::%make-cell)
(define-standard-primop-generator cell-value xlscmp::%cell-value)
(define-standard-primop-generator set-cell-value xlscmp::%set-cell-value)

(define-standard-primop-generator not xlscmp::%not)
(define-standard-primop-generator consp xlscmp::%consp)
(define-standard-primop-generator endp xlscmp::%endp)
(define-standard-primop-generator eq xlscmp::%eq)
(define-standard-primop-generator eql xlscmp::%eql)
(define-standard-primop-generator equal xlscmp::%equal)

(define-standard-primop-generator elt xlscmp::%elt)
(define-standard-primop-generator xlscmp::%set-elt)

(define-simple-primop-generator slot-value (a v)
  (case (length a)
    (1 `(xlscmp::%slot-value ,(first a) ,v))
    (2 `(xlscmp::%set-slot-value ,(first a) ,(second a) ,v))))


;;;;
;;;; Testop Code Generators
;;;;

(let ((table (make-hash-table)))
  (defun get-testop-generator (op) (gethash op table))
  (defun set-testop-generator (op g) (setf (gethash op table) g)))

(defun gencode-testop (e regs env cl)
  (let ((g (get-testop-generator (testop-fun e))))
    (unless g (error "can't generate code for testop ~s" (testop-fun e)))
    (funcall g e regs env cl)))

(defmacro define-testop-generator (which args &rest body)
  `(progn
     (set-testop-generator ',which #'(lambda ,args ,@body))
     ',which))

(defmacro define-simple-testop-generator (name symbol index)
  `(define-testop-generator ,name (e regs env cl)
     (let ((a (mapcar #'(lambda (x) (find-register x regs)) (testop-args e))))
       (multiple-value-bind (labels label-conts)
                            (make-testop-labels-conts (testop-conts e)
                                                      regs env)
         (let ((tvar (first labels))
               (evar (second labels)))
           (push-instruction `(,',symbol ,,index ,tvar ,evar ,@a) cl)
           (dolist (lc label-conts)
             (push-label-instruction (first lc) cl)
             (gencode-cexpr (second lc) regs env cl)))))))

(defun make-testop-labels-conts (conts regs env)
  (let ((labels nil)
        (rlc nil))
    (dolist (c conts (values (nreverse labels) (nreverse rlc)))
      (cond
       ((and (call-p c)
             (not (call-is-mvcall c))
             (null (call-form c))
             (let ((def (cdr (assoc (call-fun c) env))))
               (and def
                    (every #'(lambda (x y)
                               (= (find-register x regs)
                                  (find-register y regs)))
                           (call-args c)
                           (func-vars def)))))
        (push (call-fun c) labels))
        (t (let ((lab (make-label-var (gensym 'test))))
             (push lab labels)
             (push (list lab c) rlc)))))))

(define-simple-testop-generator t xlscmp::%test-1 0)
(define-simple-testop-generator consp xlscmp::%test-1 1)
(define-simple-testop-generator xlscmp::%supplied-p xlscmp::%test-1 2)
(define-simple-testop-generator endp xlscmp::%test-1 3)

(define-simple-testop-generator eq xlscmp::%test-2 0)
(define-simple-testop-generator eql xlscmp::%test-2 1)
(define-simple-testop-generator equal xlscmp::%test-2 2)

(define-simple-testop-generator = xlscmp::%test-arith-2 (char-int #\=))
(define-simple-testop-generator /= xlscmp::%test-arith-2 (char-int #\#))
(define-simple-testop-generator < xlscmp::%test-arith-2 (char-int #\<))
(define-simple-testop-generator > xlscmp::%test-arith-2 (char-int #\>))
(define-simple-testop-generator <= xlscmp::%test-arith-2 (char-int #\L))
(define-simple-testop-generator >= xlscmp::%test-arith-2 (char-int #\G))

(define-testop-generator case (e regs env cl)
  (let ((ci (find-literal (testop-data e) cl))
        (r (find-register (first (testop-args e)) regs))
        (conts (testop-conts e)))
    (multiple-value-bind (labels label-conts)
                         (make-testop-labels-conts conts regs env)
      (push-instruction `(xlscmp::%ncase ,r ,ci ,@labels) cl)
      (dolist (lc label-conts)
        (push-label-instruction (first lc) cl)
        (gencode-cexpr (second lc) regs env cl)))))


;;;;
;;;; Letrec Code Generators
;;;;

(defun gencode-letrec (e regs env cl)
  (let ((vars (letrec-vars e))
        (args (letrec-args e)))
    (do ((vars vars (rest vars))
         (args args (rest args)))
        ((or (null vars) (null args)))
        (push (cons (first vars) (first args)) env))
    (gencode-cexpr (letrec-body e) regs env cl)
    (do ((vars vars (rest vars))
         (args args (rest args)))
        ((or (null vars) (null args)))
        (push-label-instruction (first vars) cl)
        (gencode-cexpr (first args) regs env cl))))



;;;;
;;;; Call Code Generators
;;;;

(defun gencode-call-1 (e regs env cl f csym scsym mcsym smcsym)
  (let* ((form (call-form e))
         (args (call-args e))
         (na (if (consp args) (length args) 0))
         (v (let ((vars (call-vars e)))
              (if (and (consp vars) (= (length vars) 1))
                  (find-register (first vars) regs)
                0)))
         (a (if (consp args)
                (mapcar #'(lambda (x) (find-register x regs)) args)
              nil)))
    (if (null form)
        (if (call-is-mvcall e)
            (push-instruction `(,mcsym ,f 0) cl)
          (push-instruction `(,csym ,f 0 ,na ,@a) cl))
      (progn
        (if (call-is-mvcall e)
            (push-instruction `(,smcsym ,f ,v) cl)
          (push-instruction `(,scsym ,f ,v ,na ,@a) cl))
        (gencode-cexpr form regs env cl)))))

(defun gencode-call (e regs env cl)
  (let ((fun (call-fun e)))
    (cond
     ((symbolp fun)
      (gencode-call-1 e regs env cl (find-literal fun cl)
                      'xlscmp::%call 'xlscmp::%save-call
                      'xlscmp::%mvcall 'xlscmp::%save-mvcall))
     ((const-p fun)
      (let ((val (const-value fun)))
        (unless (or (symbolp val) (functionp val))
                (error "~s is not a valid function" val))
        (gencode-call-1 e regs env cl (find-literal val cl)
                        'xlscmp::%call 'xlscmp::%save-call
                        'xlscmp::%mvcall 'xlscmp::%save-mvcall)))
     ((and (var-p fun) (null (assoc fun env)))
      (gencode-call-1 e regs env cl (find-register fun regs)
                      'xlscmp::%vcall 'xlscmp::%save-vcall
                      'xlscmp::%mvvcall 'xlscmp::%save-mvvcall))
     ((and (var-p fun) (eq (func-type (cdr (assoc fun env))) 'local-lambda))
      (gencode-call-1 e regs env cl fun nil 'xlscmp::%save-lcall nil nil))
     ((and (var-p fun) (eq (func-type (cdr (assoc fun env))) 'tail-lambda))
      (let* ((fval (cdr (assoc fun env)))
             (aregs (mapcar #'(lambda (x) (find-register x regs))
                            (call-args e)))
             (vregs (mapcar #'(lambda (x) (find-register x regs))
                            (func-vars fval))))
        (gencode-move aregs vregs cl)
        (push-instruction `(xlscmp::%goto ,fun) cl)))
     (t (error "don't know how to handle the call ~s" e)))))

(defun gencode-move-copy (pairs from cl)
  (let ((npairs nil)
        (changed nil))
    (dolist (p pairs)
      (cond
       ((member (cdr p) from) (push p npairs))
       (t (push-instruction `(xlscmp::%copy ,(car p) ,(cdr p)) cl)
          (setf changed t))))
    (if changed
        (gencode-move-copy npairs (mapcar #'first npairs) cl)
      npairs)))

(defun gencode-move-swap (pairs cl)
  (do ((pairs pairs (rest pairs)))
      ((null pairs))
      (let ((p (first pairs)))
        (unless (= (car p) (cdr p))
                (let ((q (find (cdr p) pairs :key #'car)))
                  (unless (null (rest pairs))
                          (setf (car q) (car p)))
                  (push-instruction `(xlscmp::%swap ,(car p) ,(cdr p)) cl))))))

(defun gencode-move (from to cl)
  (let* ((pairs (mapcar #'cons from to))
         (spairs (gencode-move-copy pairs from cl)))
    (gencode-move-swap spairs cl)))

(defun cmpl1 (e)
  (multiple-value-bind (code lits)
                       (gencode e)
    (let* ((labels (remove-if-not #'var-p code))
           (pairs (mapcar #'(lambda (x) (cons x (gensym (var-name x))))
                          labels))
           (c (sublis pairs code)))
      (list c lits (first c)))))

(defun cmpl (e) (byte-code-close (xlscmp::assemble (cmpl1 e))))

(defun pcmpl (e) (pprint (cmpl1 e)))
