(in-package "NEWCMP")

;;;;
;;;; Data Representation
;;;;


;;;
;;; Global Variables
;;;

(defparameter *cexpr-indent-level* 0)
(defparameter *cexpr-indent-step* 1)


;;;
;;; Values
;;;

(defstruct val type name)

(defstruct (const (:include val)
                  (:constructor (new-const (value type name)))
                  (:print-function
                   (lambda (x s d)
                     (declare (ignore d))
                     (let ((n (const-name x)))
                       (if n
                           (prin1 n s)
                         (let  ((v (const-value x)))
                           (prin1 (if (constantp v) v `(quote ,v)) s)))))))
  value)

(defun make-const (value &optional name)
  (new-const value `(eql ,value) name))

(defstruct (var (:include val)
                (:constructor (make-var (name type context)))
                (:print-function (lambda (x s d)
                                     (declare (ignore d))
                                     (prin1 (var-name x) s))))
  context tag)


;;;
;;; Compound Expressions
;;;

(defstruct cexpr fun args vars conts context)


;;
;; Primop Expressions
;;

(defstruct (primop (:include cexpr)
                   (:constructor
                    (new-primop (fun args vars conts data context)))
                   (:print-function print-primop))
  data)

(defun make-primop (fun args vars form data context)
  (new-primop fun args vars (if form (list form)) data context))

(defun primop-form (e) (first (primop-conts e)))

(defun print-primop (e s d)
  (declare (ignore d))
  (let* ((*cexpr-indent-level* (+ *cexpr-indent-level* *cexpr-indent-step*))
         (fun (primop-fun e))
         (args (primop-args e))
         (vars (primop-vars e))
         (form (primop-form e))
         (data (primop-data e))
         (fd (if data (list fun data) fun))
         (lev *cexpr-indent-level*))
    (cond
     (form (format s "(PRIMOP ~s ~s ~s ~%~vt~s)" fd args vars lev form))
     (vars (format s "(PRIMOP ~s ~s ~s)" fd args vars))
     (t (format s "(PRIMOP ~s ~s)" fd args)))))

(defun primop-eq (e fun)
  (and (primop-p e) (eq (primop-fun e) fun)))


;;
;; Function Expressions
;;

(defstruct (func
            (:include cexpr)
            (:constructor
             (new-func (vars name conts type nreq nopt has-rest context)))
            (:print-function print-func))
  name type nreq nopt has-rest (nfree 0))

(defun make-func (name lambda-list body type context)
  (let* ((vars (remove '&optional (remove '&rest lambda-list)))
         (nv (length vars))
         (opt-pos (position '&optional lambda-list))
         (has-rest (if (member '&rest lambda-list) t nil))
         (nreq (if has-rest
                   (if opt-pos opt-pos (- nv 1))
                 (if opt-pos opt-pos nv)))
         (nopt (- nv nreq (if has-rest 1 0))))
    (new-func vars name (list body) type nreq nopt has-rest context)))

(defun func-body (e) (first (func-conts e)))

(defun func-lambda-list (e)
  (let* ((vars (func-vars e))
         (nreq (func-nreq e))
         (has-rest (func-has-rest e))
         (rvrest (nthcdr nreq vars))
         (reqvars (ldiff vars rvrest))
         (optvars (if has-rest (butlast rvrest) rvrest))
         (restvar (if has-rest (first (last vars)) nil)))
    (cond
     ((and has-rest optvars)
      `(,@reqvars &optional ,@optvars &rest ,restvar))
     (has-rest `(,@reqvars &rest ,restvar))
     (optvars `(,@reqvars &optional ,@optvars))
     (t reqvars))))

(defun print-func (e s d)
  (declare (ignore d))
  (let ((*cexpr-indent-level* (+ *cexpr-indent-level* *cexpr-indent-step*)))
    (format s "(~s ~s ~s~%~vt~s)"
            (func-type e)
            (func-name e)
            (func-lambda-list e)
            *cexpr-indent-level*
            (func-body e))))


;;
;; Call Expressions
;;

(defstruct (call (:include cexpr)
                 (:constructor (new-call (fun args vars conts val-ignored
                                              is-mvcall context)))
                 (:print-function print-call))
  val-ignored is-mvcall (weight 1))

(defun make-call (fun args vars form val-ignored is-mvcall context)
  (new-call fun args vars (if form (list form) nil) val-ignored
            is-mvcall context))

(defun call-form (e) (first (call-conts e)))

(defun print-call (e s d)
  (declare (ignore d))
  (let* ((*cexpr-indent-level* (+ *cexpr-indent-level* *cexpr-indent-step*))
         (fun (call-fun e))
         (args (call-args e))
         (pargs (if (call-is-mvcall e) t args))
         (vars (call-vars e))
         (form (call-form e))
         (lev *cexpr-indent-level*))
    (if form
        (let ((pvars (if (and (null vars) (not (call-val-ignored e)))
                         t
                       vars)))
          (format s "(CALL ~s ~s ~s~%~vt~s)" fun pargs pvars lev form))
      (format s "(CALL ~s ~s)" fun pargs))))


;;
;; Binding Expressions
;;

(defstruct (bind (:include cexpr)
                 (:constructor (new-bind (vars args conts context)))
                 (:print-function print-bindings)))

(defstruct (letrec (:include cexpr)
                   (:constructor (new-letrec (vars args conts context)))
                   (:print-function print-bindings)))

(defun make-bind (vars args body context)
  (new-bind vars args (list body) context))

(defun make-letrec (vars args body context)
  (new-letrec vars args (list body) context))

(defun bind-body (e) (first (bind-conts e)))

(defun letrec-body (e) (first (letrec-conts e)))

(defun print-bindings (x s d)
  (declare (ignore d))
  (let* ((head (if (bind-p x) "(BIND (" "(LETREC ("))
         (vars (letrec-vars x))
         (args (letrec-args x))
         (body (letrec-body x))
         (vs (format nil "~s" (first vars))))
    (format s "~a(~a " head vs)
    (let ((baselev (+ *cexpr-indent-level* (length head))))
      (let ((*cexpr-indent-level* (+ baselev (length vs) 2)))
        (format s "~s)" (first args)))
      (mapc #'(lambda (v a)
                (let ((vs (format nil "~s" v)))
                  (format s "~%~vt(~a " baselev vs)
                  (let ((*cexpr-indent-level* (+ baselev (length vs) 2)))
                    (format s "~s)" a))))
            (rest vars)
            (rest args)))
    (format s ")")
    (let ((*cexpr-indent-level* (+ *cexpr-indent-level* *cexpr-indent-step*)))
      (format s "~%~vt~s)" *cexpr-indent-level* body))))


;;
;; Testop Expressions
;;

(defstruct (testop (:include cexpr)
                   (:constructor (make-testop (fun args conts data context)))
                   (:print-function print-testop))
  data)

(defun print-testop (e s d)
  (declare (ignore d))
  (let* ((fun (testop-fun e))
         (args (testop-args e))
         (conts (testop-conts e))
         (data (testop-data e))
         (fd (if data (list fun data) fun)))
    (format s "(TESTOP ~s ~s" fd args)
    (let ((*cexpr-indent-level*
           (+ *cexpr-indent-level* *cexpr-indent-step* 1)))
      (format s "~%~vt(~s" (- *cexpr-indent-level* 1) (first conts))
      (dolist (c (rest conts))
        (format s "~%~vt~s" *cexpr-indent-level* c))
      (format s "))"))))


;;;
;;; Utility Functions
;;;

(defun remake-cexpr (e fun args vars conts)
  (flet ((eqlist (a b)
           (or (eq a b) (and (= (length a) (length b)) (every #'eq a b)))))
    (if (and (eq fun (cexpr-fun e))
             (eqlist args (cexpr-args e))
             (eqlist vars (cexpr-vars e))
             (eqlist conts (cexpr-conts e)))
        e
      (let ((ne (copy-structure e)))
        (setf (cexpr-fun ne) fun)
        (setf (cexpr-args ne) args)
        (setf (cexpr-vars ne) vars)
        (setf (cexpr-conts ne) conts)
        ne))))


;;;;
;;;; Parser for Intermediate Form
;;;;

(defun parse-value (e &optional env)
  (cond
   ((and (symbolp e) (constantp e))
    (make-const (symbol-value e) e))
   ((and (consp e) (eq (first e) 'quote) (null (rest (rest e))))
    (make-const (second e)))
   ((constantp e) (make-const e))
   ((symbolp e)
    (let ((var (rest (assoc e env))))
      (if var var (error "symbol ~s is not bound" e))))
   (t (error "bad value expression: ~s" e))))

(defun parse-cexpr (e &optional env)
  (if (consp e)
      (case (first e)
            (primop (parse-primop e env))
            ((lambda local-lambda tail-lambda) (parse-func e env))
            (call (parse-call e env))
            ((bind letrec) (parse-binding e env))
            (testop (parse-testop e env))
            (t (error "bad compound expression: ~s" e)))
    (error "bad compound expression: ~s" e)))

(defun parse-cexpr-args (args env &optional t-ok)
  (if (and t-ok (eq args t))
      t
    (mapcar #'(lambda (x) (parse-value x env)) args)))

(defun parse-cexpr-vars (vars env &optional t-ok)
  (if (and t-ok (eq vars t))
      (values t env)
    (let ((nll nil))
      (dolist (s vars)
        (if (member s '(&optional &rest))
            (push s nll)
          (let ((v (make-var s t nil)))
            (push (cons s v) env)
            (push v nll))))
      (values (nreverse nll) env))))

(defun parse-primop (e &optional env)
  (let* ((fd (second e))
         (fun (if (consp fd) (first fd) fd))
         (data (if (consp fd) (second fd) nil))
         (args (parse-cexpr-args (third e) env t)))
    (case (length e)
          (3 (make-primop fun args nil nil data nil))
          ((4 5)
           (let ((vars (fourth e))
                 (form (fifth e)))
             (multiple-value-bind (parsed-vars new-env)
                                  (parse-cexpr-vars vars env t)
               (let ((parsed-form (if form (parse-cexpr form new-env) nil)))
                 (make-primop fun args parsed-vars parsed-form data nil)))))
          (t (error "bad primop form: ~s" e)))))        

(defun parse-func (e &optional env)
  (let ((name (second e))
        (llist (third e))
        (body (fourth e)))
    (multiple-value-bind (new-llist new-env)
                         (parse-cexpr-vars llist env)
      (make-func name new-llist (parse-cexpr body new-env) (first e) nil))))

(defun parse-call (e &optional env)
  (let* ((fun (parse-call-fun (second e) env))
         (is-mvcall (eq (third e) t))
         (args (if is-mvcall nil (parse-cexpr-args (third e) env t))))
    (case (length e)
          (3 (make-call fun args nil nil nil is-mvcall nil))
          (5 (multiple-value-bind (parsed-vars new-env)
                                  (parse-cexpr-vars (fourth e) env t)
               (let* ((parsed-form (parse-cexpr (fifth e) new-env))
                      (vars (if (eq parsed-vars t) nil parsed-vars))
                      (vi (if (and parsed-form (null parsed-vars)) t nil)))
                 (make-call fun args vars parsed-form vi is-mvcall nil)))))))

(defun parse-call-fun (e env)
  (unless (symbolp e) (error "bad call function: ~s" e))
  (let ((var (rest (assoc e env))))
    (if var var e)))

(defun parse-binding (e env)
  (let ((vars (mapcar #'first (second e)))
        (args (mapcar #'second (second e))))
    (multiple-value-bind (parsed-vars new-env)
                         (parse-cexpr-vars vars env)
      (let ((parsed-args (parse-letrec-args args new-env))
            (parsed-form (parse-cexpr (third e) new-env)))
        (if (eq (first e) 'bind)
            (make-bind parsed-vars parsed-args parsed-form nil)
          (make-letrec parsed-vars parsed-args parsed-form nil))))))

(defun parse-letrec-args (args env)
  (mapcar #'(lambda (e) (parse-func e env)) args))

(defun parse-testop (e &optional env)
  (let* ((fd (second e))
         (fun (if (consp fd) (first fd) fd))
         (data (if (consp fd) (second fd) nil))
         (args (parse-cexpr-args (third e) env t))
         (forms (fourth e))
         (parsed-forms (mapcar #'(lambda (x) (parse-cexpr x env)) forms)))
    (make-testop fun args parsed-forms data nil)))




;;;;
;;;; Some Simple Transformations
;;;;

;;;
;;; Value Substitution
;;;

(defun substitute-values (e &optional env)
  (labels ((sv (e env)
             (if (bind-p e)
                 (let* ((args (sublis env (cexpr-args e)))
                        (vars (cexpr-vars e))
                        (nenv (append (mapcar #'cons vars args) env)))
                   (substitute-values (letrec-body e) nenv))
               (let ((fun (sublis env (cexpr-fun e)))
                     (args (if (letrec-p e)
                               (mapcar #'(lambda (x) (sv x env)) (cexpr-args e))
                             (sublis env (cexpr-args e))))
                     (vars (cexpr-vars e))
                     (conts (mapcar #'(lambda (x) (sv x env)) (cexpr-conts e))))
                 (remake-cexpr e fun args vars conts)))))
    (sv e env)))



;;;
;;; Cell Insertion
;;;

(defun find-setq-vars (e vars)
  (cond
   ((primop-eq e 'setq)
    (find-setq-vars (primop-form e) (cons (first (primop-args e)) vars)))
   ((letrec-p e)
    (let ((nvars (find-setq-vars-list (cexpr-args e) vars)))
      (find-setq-vars-list (cexpr-conts e) nvars)))
   ((cexpr-p e) (find-setq-vars-list (cexpr-conts e) vars))
   (t vars)))

(defun find-setq-vars-list (elist vars)
  (if elist
      (find-setq-vars-list (rest elist) (find-setq-vars (first elist) vars))
    vars))
(defun make-cell-env (e)
  (mapcar #'(lambda (v)
              (let ((cv (copy-var v)))
                (setf (var-type cv) `(cell ,(var-type v)))
                (cons v cv)))
          (find-setq-vars e nil)))

(defun add-cells (e)
  (labels ((ac (e env)
             (cond
              ((primop-eq e 'setq)
               (let* ((args (primop-args e))
                      (vars (primop-vars e))
                      (form (primop-form e))
                      (cell (rest (assoc (first args) env)))
                      (nargs (cons cell (rest args)))
                      (ctx (primop-context e))
                      (ne (make-primop 'set-cell-value nargs vars form nil ctx)))
                 (ac ne env)))
              ((letrec-p e)
               (let ((args (mapcar #'(lambda (x) (ac x env)) (letrec-args e)))
                     (vars (letrec-vars e))
                     (body (ac (letrec-body e) env)))
                 (remake-cexpr e nil args vars (list body))))
              ((cexpr-p e)
               (let* ((conts (mapcar #'(lambda (x) (ac x env)) (cexpr-conts e)))
                        (vars (cexpr-vars e))
                        (nconts (if (consp vars)
                                    (wrap-new-cells vars (first conts) env)
                                  conts)))
                   (add-ref-cells e nconts env))))))
    (ac e (make-cell-env e))))

(defun wrap-new-cells (vars cont env)
  (dolist (v vars (list cont))
    (let ((c (cdr (assoc v env))))
      (when c
            (setf cont (make-primop 'make-cell (list v) (list c) cont
                                    nil (var-context v)))))))

(defun add-ref-cells (e conts env)
  (let ((fun (cexpr-fun e))
        (vars (cexpr-vars e))
        (args (cexpr-args e)))
    (if (consp args)
        (let* ((pairs (let ((pairs nil))
                        (dolist (a args pairs)
                          (when (assoc a env)
                                (push (cons a (copy-var a)) pairs)))))
               (nargs (sublis pairs args))
               (ne (remake-cexpr e fun nargs vars conts)))
          (dolist (p pairs ne)
            (let ((c (cdr (assoc (first p) env)))
                  (v (cdr p)))
              (setf ne (make-primop 'cell-value (list c) (list v) ne
                                    nil (var-context c))))))
      (remake-cexpr e fun args vars conts))))
